Mercurial > public > mercurial-scm > hg
comparison mercurial/util.py @ 1241:3b4f05ff3130
Add support for cloning with hardlinks on windows.
In order to use hardlinks, the win32file module is needed, and this is
present in ActivePython. If it isn't present, or hardlinks are not supported
on the underlying filesystem, a regular copy is used.
When using hardlinks the biggest benefit is probably the saving in space,
but cloning can be much quicker. For example cloning the Xen tree
(non trivial) without an update goes from about 95s to 15s.
Unix-like platforms should be unaffected, although should be more tolerant on
filesystems that don't support hard links.
(tweaked by mpm to deal with new copyfiles function)
--- hg.orig/mercurial/commands.py 2005-09-13 19:32:53.000000000 -0500
+++ hg/mercurial/commands.py 2005-09-14 12:11:34.000000000 -0500
@@ -620,10 +620,6 @@ def clone(ui, source, dest=None, **opts)
if other.dev() != -1:
abspath = os.path.abspath(source)
- copyfile = (os.stat(dest).st_dev == other.dev()
- and getattr(os, 'link', None) or shutil.copy2)
- if copyfile is not shutil.copy2:
- ui.note("cloning by hardlink\n")
# we use a lock here because if we race with commit, we can
# end up with extra data in the cloned revlogs that's not
@@ -638,7 +634,7 @@ def clone(ui, source, dest=None, **opts)
for f in files.split():
src = os.path.join(source, ".hg", f)
dst = os.path.join(dest, ".hg", f)
- util.copyfiles(src, dst, copyfile)
+ util.copyfiles(src, dst)
repo = hg.repository(ui, dest)
Index: hg/mercurial/util.py
===================================================================
--- hg.orig/mercurial/util.py 2005-09-08 00:15:25.000000000 -0500
+++ hg/mercurial/util.py 2005-09-14 12:16:49.000000000 -0500
@@ -12,7 +12,7 @@ platform-specific details from the core.
import os, errno
from demandload import *
-demandload(globals(), "re cStringIO")
+demandload(globals(), "re cStringIO shutil")
def binary(s):
"""return true if a string is binary data using diff's heuristic"""
@@ -217,17 +217,28 @@ def rename(src, dst):
os.unlink(dst)
os.rename(src, dst)
-def copyfiles(src, dst, copyfile):
- """Copy a directory tree, files are copied using 'copyfile'."""
+def copyfiles(src, dst, hardlink=None):
+ """Copy a directory tree using hardlinks if possible"""
+
+ if hardlink is None:
+ hardlink = (os.stat(src).st_dev ==
+ os.stat(os.path.dirname(dst)).st_dev)
if os.path.isdir(src):
os.mkdir(dst)
for name in os.listdir(src):
srcname = os.path.join(src, name)
dstname = os.path.join(dst, name)
- copyfiles(srcname, dstname, copyfile)
+ copyfiles(srcname, dstname, hardlink)
else:
- copyfile(src, dst)
+ if hardlink:
+ try:
+ os_link(src, dst)
+ except:
+ hardlink = False
+ shutil.copy2(src, dst)
+ else:
+ shutil.copy2(src, dst)
def opener(base):
"""
@@ -244,13 +255,13 @@ def opener(base):
if mode[0] != "r":
try:
- s = os.stat(f)
+ nlink = nlinks(f)
except OSError:
d = os.path.dirname(f)
if not os.path.isdir(d):
os.makedirs(d)
else:
- if s.st_nlink > 1:
+ if nlink > 1:
file(f + ".tmp", "wb").write(file(f, "rb").read())
rename(f+".tmp", f)
@@ -266,10 +277,41 @@ def _makelock_file(info, pathname):
def _readlock_file(pathname):
return file(pathname).read()
+def nlinks(pathname):
+ """Return number of hardlinks for the given file."""
+ return os.stat(pathname).st_nlink
+
+if hasattr(os, 'link'):
+ os_link = os.link
+else:
+ def os_link(src, dst):
+ raise OSError(0, "Hardlinks not supported")
+
# Platform specific variants
if os.name == 'nt':
nulldev = 'NUL:'
+ try: # ActivePython can create hard links using win32file module
+ import win32file
+
+ def os_link(src, dst): # NB will only succeed on NTFS
+ win32file.CreateHardLink(dst, src)
+
+ def nlinks(pathname):
+ """Return number of hardlinks for the given file."""
+ try:
+ fh = win32file.CreateFile(pathname,
+ win32file.GENERIC_READ, win32file.FILE_SHARE_READ,
+ None, win32file.OPEN_EXISTING, 0, None)
+ res = win32file.GetFileInformationByHandle(fh)
+ fh.Close()
+ return res[7]
+ except:
+ return os.stat(pathname).st_nlink
+
+ except ImportError:
+ pass
+
def is_exec(f, last):
return last
author | Stephen Darnell |
---|---|
date | Wed, 14 Sep 2005 12:22:20 -0500 |
parents | a7b8812973d9 |
children | 1945754e466b |
comparison
equal
deleted
inserted
replaced
1240:cc756ffd4d04 | 1241:3b4f05ff3130 |
---|---|
10 platform-specific details from the core. | 10 platform-specific details from the core. |
11 """ | 11 """ |
12 | 12 |
13 import os, errno | 13 import os, errno |
14 from demandload import * | 14 from demandload import * |
15 demandload(globals(), "re cStringIO") | 15 demandload(globals(), "re cStringIO shutil") |
16 | 16 |
17 def binary(s): | 17 def binary(s): |
18 """return true if a string is binary data using diff's heuristic""" | 18 """return true if a string is binary data using diff's heuristic""" |
19 if s and '\0' in s[:4096]: | 19 if s and '\0' in s[:4096]: |
20 return True | 20 return True |
215 os.rename(src, dst) | 215 os.rename(src, dst) |
216 except: | 216 except: |
217 os.unlink(dst) | 217 os.unlink(dst) |
218 os.rename(src, dst) | 218 os.rename(src, dst) |
219 | 219 |
220 def copyfiles(src, dst, copyfile): | 220 def copyfiles(src, dst, hardlink=None): |
221 """Copy a directory tree, files are copied using 'copyfile'.""" | 221 """Copy a directory tree using hardlinks if possible""" |
222 | |
223 if hardlink is None: | |
224 hardlink = (os.stat(src).st_dev == | |
225 os.stat(os.path.dirname(dst)).st_dev) | |
222 | 226 |
223 if os.path.isdir(src): | 227 if os.path.isdir(src): |
224 os.mkdir(dst) | 228 os.mkdir(dst) |
225 for name in os.listdir(src): | 229 for name in os.listdir(src): |
226 srcname = os.path.join(src, name) | 230 srcname = os.path.join(src, name) |
227 dstname = os.path.join(dst, name) | 231 dstname = os.path.join(dst, name) |
228 copyfiles(srcname, dstname, copyfile) | 232 copyfiles(srcname, dstname, hardlink) |
229 else: | 233 else: |
230 copyfile(src, dst) | 234 if hardlink: |
235 try: | |
236 os_link(src, dst) | |
237 except: | |
238 hardlink = False | |
239 shutil.copy2(src, dst) | |
240 else: | |
241 shutil.copy2(src, dst) | |
231 | 242 |
232 def opener(base): | 243 def opener(base): |
233 """ | 244 """ |
234 return a function that opens files relative to base | 245 return a function that opens files relative to base |
235 | 246 |
242 | 253 |
243 mode += "b" # for that other OS | 254 mode += "b" # for that other OS |
244 | 255 |
245 if mode[0] != "r": | 256 if mode[0] != "r": |
246 try: | 257 try: |
247 s = os.stat(f) | 258 nlink = nlinks(f) |
248 except OSError: | 259 except OSError: |
249 d = os.path.dirname(f) | 260 d = os.path.dirname(f) |
250 if not os.path.isdir(d): | 261 if not os.path.isdir(d): |
251 os.makedirs(d) | 262 os.makedirs(d) |
252 else: | 263 else: |
253 if s.st_nlink > 1: | 264 if nlink > 1: |
254 file(f + ".tmp", "wb").write(file(f, "rb").read()) | 265 file(f + ".tmp", "wb").write(file(f, "rb").read()) |
255 rename(f+".tmp", f) | 266 rename(f+".tmp", f) |
256 | 267 |
257 return file(f, mode) | 268 return file(f, mode) |
258 | 269 |
264 os.close(ld) | 275 os.close(ld) |
265 | 276 |
266 def _readlock_file(pathname): | 277 def _readlock_file(pathname): |
267 return file(pathname).read() | 278 return file(pathname).read() |
268 | 279 |
280 def nlinks(pathname): | |
281 """Return number of hardlinks for the given file.""" | |
282 return os.stat(pathname).st_nlink | |
283 | |
284 if hasattr(os, 'link'): | |
285 os_link = os.link | |
286 else: | |
287 def os_link(src, dst): | |
288 raise OSError(0, "Hardlinks not supported") | |
289 | |
269 # Platform specific variants | 290 # Platform specific variants |
270 if os.name == 'nt': | 291 if os.name == 'nt': |
271 nulldev = 'NUL:' | 292 nulldev = 'NUL:' |
293 | |
294 try: # ActivePython can create hard links using win32file module | |
295 import win32file | |
296 | |
297 def os_link(src, dst): # NB will only succeed on NTFS | |
298 win32file.CreateHardLink(dst, src) | |
299 | |
300 def nlinks(pathname): | |
301 """Return number of hardlinks for the given file.""" | |
302 try: | |
303 fh = win32file.CreateFile(pathname, | |
304 win32file.GENERIC_READ, win32file.FILE_SHARE_READ, | |
305 None, win32file.OPEN_EXISTING, 0, None) | |
306 res = win32file.GetFileInformationByHandle(fh) | |
307 fh.Close() | |
308 return res[7] | |
309 except: | |
310 return os.stat(pathname).st_nlink | |
311 | |
312 except ImportError: | |
313 pass | |
272 | 314 |
273 def is_exec(f, last): | 315 def is_exec(f, last): |
274 return last | 316 return last |
275 | 317 |
276 def set_exec(f, mode): | 318 def set_exec(f, mode): |